@@ -1,23 +1,52 @@
+# This file is part of the TREZOR project.
+#
+# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
+# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
+# Copyright (C) 2016      Jochen Hoenicke <hoenicke@gmail.com>
+#
+# This library is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Lesser General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This library is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Lesser General Public License for more details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with this library.  If not, see <http://www.gnu.org/licenses/>.
+
 import struct
 import hmac
 import hashlib
+import sys
 
 import ecdsa
 from ecdsa.util import string_to_number, number_to_string
 from ecdsa.curves import SECP256k1
 from ecdsa.ellipticcurve import Point, INFINITY
 
-import tools
-import types_pb2 as proto_types
+from . import tools
+from . import types_pb2 as proto_types
 
 PRIME_DERIVATION_FLAG = 0x80000000
 
+if sys.version_info < (3,):
+    def byteindex(data, index):
+        return ord(data[index])
+else:
+    def byteindex(data, index):
+        return data[index]
+
+
 def point_to_pubkey(point):
     order = SECP256k1.order
     x_str = number_to_string(point.x(), order)
     y_str = number_to_string(point.y(), order)
     vk = x_str + y_str
-    return chr((ord(vk[63]) & 1) + 2) + vk[0:32]  # To compressed key
+    return struct.pack('B', (byteindex(vk, 63) & 1) + 2) + vk[0:32]  # To compressed key
+
 
 def sec_to_public_pair(pubkey):
     """Convert a public key in sec binary format to a public pair."""
@@ -37,15 +66,19 @@ def sec_to_public_pair(pubkey):
 
     return public_pair_for_x(ecdsa.ecdsa.generator_secp256k1, x, is_even=(sec0 == b'\2'))
 
+
 def is_prime(n):
     return (bool)(n & PRIME_DERIVATION_FLAG)
 
+
 def fingerprint(pubkey):
     return string_to_number(tools.hash_160(pubkey)[:4])
 
+
 def get_address(public_node, address_type):
     return tools.public_key_to_bc_address(public_node.public_key, address_type)
 
+
 def public_ckd(public_node, n):
     if not isinstance(n, list):
         raise Exception('Parameter must be a list')
@@ -58,6 +91,7 @@ def public_ckd(public_node, n):
 
     return node
 
+
 def get_subnode(node, i):
     # Public Child key derivation (CKD) algorithm of BIP32
     i_as_bytes = struct.pack(">L", i)
@@ -79,8 +113,7 @@ def get_subnode(node, i):
 
     # BIP32 magic converts old public key to new public point
     x, y = sec_to_public_pair(node.public_key)
-    point = I_left_as_exponent * SECP256k1.generator + \
-            Point(SECP256k1.curve, x, y, SECP256k1.order)
+    point = I_left_as_exponent * SECP256k1.generator + Point(SECP256k1.curve, x, y, SECP256k1.order)
 
     if point == INFINITY:
         raise Exception("Point cannot be INFINITY")
@@ -90,20 +123,22 @@ def get_subnode(node, i):
 
     return node_out
 
+
 def serialize(node, version=0x0488B21E):
-    s = ''
+    s = b''
     s += struct.pack('>I', version)
     s += struct.pack('>B', node.depth)
     s += struct.pack('>I', node.fingerprint)
     s += struct.pack('>I', node.child_num)
     s += node.chain_code
     if node.private_key:
-        s += '\x00' + node.private_key
-    else :
+        s += b'\x00' + node.private_key
+    else:
         s += node.public_key
     s += tools.Hash(s)[:4]
     return tools.b58encode(s)
 
+
 def deserialize(xpub):
     data = tools.b58decode(xpub, None)
 
@@ -117,7 +152,7 @@ def deserialize(xpub):
     node.chain_code = data[13:45]
 
     key = data[45:-4]
-    if key[0] == '\x00':
+    if byteindex(key, 0) == 0:
         node.private_key = key[1:]
     else:
         node.public_key = key
