@@ -1,4 +1,23 @@
-from __future__ import print_function
+# This file is part of the TREZOR project.
+#
+# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
+# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
+# Copyright (C) 2016      Jochen Hoenicke <hoenicke@gmail.com>
+#
+# This library is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Lesser General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This library is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Lesser General Public License for more details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with this library.  If not, see <http://www.gnu.org/licenses/>.
+
+from __future__ import print_function, absolute_import
 
 import os
 import sys
@@ -6,17 +25,24 @@ import time
 import binascii
 import hashlib
 import unicodedata
-import json
+# import json
 import getpass
 
 from mnemonic import Mnemonic
 
 from . import tools
-from . import mapping
+# from . import mapping
 from . import messages_pb2 as proto
 from . import types_pb2 as types
 from .debuglink import DebugLink
 
+# Python2 vs Python3
+try:
+    input = raw_input
+except NameError:
+    pass
+
+
 # try:
 #     from PIL import Image
 #     SCREENSHOT = True
@@ -30,9 +56,36 @@ DEFAULT_CURVE = 'secp256k1'
 # monkeypatching: text formatting of protobuf messages
 tools.monkeypatch_google_protobuf_text_format()
 
+
+def getch():
+    try:
+        import termios
+    except ImportError:
+        # Non-POSIX. Return msvcrt's (Windows') getch.
+        import msvcrt
+        return msvcrt.getch()
+
+    # POSIX system. Create and return a getch that manipulates the tty.
+    import sys
+    import tty
+
+    def _getch():
+        fd = sys.stdin.fileno()
+        old_settings = termios.tcgetattr(fd)
+        try:
+            tty.setraw(fd)
+            ch = sys.stdin.read(1)
+        finally:
+            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
+        return ch
+
+    return _getch()
+
+
 def get_buttonrequest_value(code):
     # Converts integer code to its string representation of ButtonRequestType
-    return [ k for k, v in types.ButtonRequestType.items() if v == code][0]
+    return [k for k, v in types.ButtonRequestType.items() if v == code][0]
+
 
 def pprint(msg):
     msg_class = msg.__class__.__name__
@@ -47,18 +100,23 @@ def pprint(msg):
     else:
         return "<%s> (%d bytes):\n%s" % (msg_class, msg_size, msg)
 
+
 def log(msg):
-    sys.stderr.write("%s\n" % msg)
+    sys.stderr.write(str(msg))
+    sys.stderr.write('\n')
     sys.stderr.flush()
 
+
 class CallException(Exception):
     def __init__(self, code, message):
         super(CallException, self).__init__()
         self.args = [code, message]
 
+
 class PinException(CallException):
     pass
 
+
 class field(object):
     # Decorator extracts single value from
     # protobuf object. If the field is not
@@ -73,6 +131,7 @@ class field(object):
             return getattr(ret, self.field)
         return wrapped_f
 
+
 class expect(object):
     # Decorator checks if the method
     # returned one of expected protobuf messages
@@ -88,20 +147,35 @@ class expect(object):
             return ret
         return wrapped_f
 
+
+def session(f):
+    # Decorator wraps a BaseClient method
+    # with session activation / deactivation
+    def wrapped_f(*args, **kwargs):
+        client = args[0]
+        try:
+            client.transport.session_begin()
+            return f(*args, **kwargs)
+        finally:
+            client.transport.session_end()
+    return wrapped_f
+
+
 def normalize_nfc(txt):
     if sys.version_info[0] < 3:
         if isinstance(txt, unicode):
-            return unicodedata.normalize('NFC', txt).encode('utf-8')
+            return unicodedata.normalize('NFC', txt)
         if isinstance(txt, str):
-            return unicodedata.normalize('NFC', txt.decode('utf-8')).encode('utf-8')
+            return unicodedata.normalize('NFC', txt.decode('utf-8'))
     else:
         if isinstance(txt, bytes):
-            return unicodedata.normalize('NFC', txt.decode('utf-8')).encode('utf-8')
+            return unicodedata.normalize('NFC', txt.decode('utf-8'))
         if isinstance(txt, str):
-            return unicodedata.normalize('NFC', txt).encode('utf-8')
+            return unicodedata.normalize('NFC', txt)
 
     raise Exception('unicode/str or bytes/str expected')
 
+
 class BaseClient(object):
     # Implements very basic layer of sending raw protobuf
     # messages to device and getting its response back.
@@ -112,39 +186,28 @@ class BaseClient(object):
     def cancel(self):
         self.transport.write(proto.Cancel())
 
+    @session
     def call_raw(self, msg):
-        try:
-            self.transport.session_begin()
-            self.transport.write(msg)
-            resp = self.transport.read_blocking()
-        finally:
-            self.transport.session_end()
-
-        return resp
+        self.transport.write(msg)
+        return self.transport.read_blocking()
 
+    @session
     def call(self, msg):
-        try:
-            self.transport.session_begin()
+        resp = self.call_raw(msg)
+        handler_name = "callback_%s" % resp.__class__.__name__
+        handler = getattr(self, handler_name, None)
 
-            resp = self.call_raw(msg)
-            handler_name = "callback_%s" % resp.__class__.__name__
-            handler = getattr(self, handler_name, None)
-
-            if handler != None:
-                msg = handler(resp)
-                if msg == None:
-                    raise Exception("Callback %s must return protobuf message, not None" % handler)
-
-                resp = self.call(msg)
-
-        finally:
-            self.transport.session_end()
+        if handler is not None:
+            msg = handler(resp)
+            if msg is None:
+                raise Exception("Callback %s must return protobuf message, not None" % handler)
+            resp = self.call(msg)
 
         return resp
 
     def callback_Failure(self, msg):
         if msg.code in (types.Failure_PinInvalid,
-            types.Failure_PinCancelled, types.Failure_PinExpected):
+                        types.Failure_PinCancelled, types.Failure_PinExpected):
             raise PinException(msg.code, msg.message)
 
         raise CallException(msg.code, msg.message)
@@ -152,6 +215,7 @@ class BaseClient(object):
     def close(self):
         self.transport.close()
 
+
 class DebugWireMixin(object):
     def call_raw(self, msg):
         log("SENDING " + pprint(msg))
@@ -159,6 +223,7 @@ class DebugWireMixin(object):
         log("RECEIVED " + pprint(resp))
         return resp
 
+
 class TextUIMixin(object):
     # This class demonstrates easy test-based UI
     # integration between the device and wallet.
@@ -173,6 +238,28 @@ class TextUIMixin(object):
         # log("Sending ButtonAck for %s " % get_buttonrequest_value(msg.code))
         return proto.ButtonAck()
 
+    def callback_RecoveryMatrix(self, msg):
+        if self.recovery_matrix_first_pass:
+            self.recovery_matrix_first_pass = False
+            log("Use the numeric keypad to describe positions.  For the word list use only left and right keys. The layout is:")
+            log("    7 8 9     7 | 9")
+            log("    4 5 6     4 | 6")
+            log("    1 2 3     1 | 3")
+        while True:
+            character = getch()
+            if character in ('\x03', '\x04'):
+                return proto.Cancel()
+
+            if character in ('\x08', '\x7f'):
+                return proto.WordAck(word='\x08')
+
+            # ignore middle column if only 6 keys requested.
+            if (isinstance(msg.type, types.WordRequestType_Matrix6) and character in ('2', '5', '8')):
+                continue
+
+            if (ord(character) >= ord('1') and ord(character) <= ord('9')):
+                return proto.WordAck(word=character)
+
     def callback_PinMatrixRequest(self, msg):
         if msg.type == 1:
             desc = 'current PIN'
@@ -203,13 +290,16 @@ class TextUIMixin(object):
             exit()
 
     def callback_WordRequest(self, msg):
+        if msg.type in (types.WordRequestType_Matrix9,
+                        types.WordRequestType_Matrix6):
+            return self.callback_RecoveryMatrix(msg)
         log("Enter one word of mnemonic: ")
-        try:
-            word = raw_input()
-        except NameError:
-            word = input() # Python 3
+        word = input()
+        if self.expand:
+            word = self.mnemonic_wordlist.expand_word(word)
         return proto.WordAck(word=word)
 
+
 class DebugLinkMixin(object):
     # This class implements automatic responses
     # and other functionality for unit tests
@@ -256,15 +346,15 @@ class DebugLinkMixin(object):
     def __exit__(self, _type, value, traceback):
         self.in_with_statement -= 1
 
-        if _type != None:
+        if _type is not None:
             # Another exception raised
             return False
 
         # return isinstance(value, TypeError)
         # Evaluate missed responses in 'with' statement
-        if self.expected_responses != None and len(self.expected_responses):
-            raise Exception("Some of expected responses didn't come from device: %s" % \
-                    [ pprint(x) for x in self.expected_responses ])
+        if self.expected_responses is not None and len(self.expected_responses):
+            raise Exception("Some of expected responses didn't come from device: %s" %
+                            [pprint(x) for x in self.expected_responses])
 
         # Cleanup
         self.expected_responses = None
@@ -283,7 +373,7 @@ class DebugLinkMixin(object):
         self.passphrase = normalize_nfc(passphrase)
 
     def set_mnemonic(self, mnemonic):
-        self.mnemonic = normalize_nfc(passphrase)
+        self.mnemonic = normalize_nfc(mnemonic).split(' ')
 
     def call_raw(self, msg):
 
@@ -304,22 +394,22 @@ class DebugLinkMixin(object):
         return resp
 
     def _check_request(self, msg):
-        if self.expected_responses != None:
+        if self.expected_responses is not None:
             try:
                 expected = self.expected_responses.pop(0)
             except IndexError:
-                raise CallException(types.Failure_Other,
-                        "Got %s, but no message has been expected" % pprint(msg))
+                raise CallException(types.Failure_UnexpectedMessage,
+                                    "Got %s, but no message has been expected" % pprint(msg))
 
             if msg.__class__ != expected.__class__:
-                raise CallException(types.Failure_Other,
-                            "Expected %s, got %s" % (pprint(expected), pprint(msg)))
+                raise CallException(types.Failure_UnexpectedMessage,
+                                    "Expected %s, got %s" % (pprint(expected), pprint(msg)))
 
             fields = expected.ListFields()  # only filled (including extensions)
             for field, value in fields:
                 if not msg.HasField(field.name) or getattr(msg, field.name) != value:
-                    raise CallException(types.Failure_Other,
-                            "Expected %s, got %s" % (pprint(expected), pprint(msg)))
+                    raise CallException(types.Failure_UnexpectedMessage,
+                                        "Expected %s, got %s" % (pprint(expected), pprint(msg)))
 
     def callback_ButtonRequest(self, msg):
         log("ButtonRequest code: " + get_buttonrequest_value(msg.code))
@@ -351,9 +441,10 @@ class DebugLinkMixin(object):
 
         raise Exception("Unexpected call")
 
+
 class ProtocolMixin(object):
     PRIME_DERIVATION_FLAG = 0x80000000
-    VENDORS = ('bitcointrezor.com',)
+    VENDORS = ('bitcointrezor.com', 'trezor.io')
 
     def __init__(self, *args, **kwargs):
         super(ProtocolMixin, self).__init__(*args, **kwargs)
@@ -373,7 +464,7 @@ class ProtocolMixin(object):
 
     def _convert_prime(self, n):
         # Convert minus signs to uint32 with flag
-        return [ int(abs(x) | self.PRIME_DERIVATION_FLAG) if x < 0 else x for x in n ]
+        return [int(abs(x) | self.PRIME_DERIVATION_FLAG) if x < 0 else x for x in n]
 
     @staticmethod
     def expand_path(n):
@@ -383,6 +474,27 @@ class ProtocolMixin(object):
             return []
 
         n = n.split('/')
+
+        # m/a/b/c => a/b/c
+        if n[0] == 'm':
+            n = n[1:]
+
+        # coin_name/a/b/c => 44'/SLIP44_constant'/a/b/c
+        coins = {
+            "Bitcoin": 0,
+            "Testnet": 1,
+            "Namecoin": 7,
+            "Litecoin": 2,
+            "Dogecoin": 3,
+            "Dash": 5,
+            "Ether": 60,
+            "EtherClassic": 61,
+            "Zcash": 133,
+            "Decred": 42
+        }
+        if n[0] in coins:
+            n = ["44'", "%d'" % coins[n[0]]] + n[1:]
+
         path = []
         for x in n:
             prime = False
@@ -402,20 +514,20 @@ class ProtocolMixin(object):
         return path
 
     @expect(proto.PublicKey)
-    def get_public_node(self, n, ecdsa_curve_name=DEFAULT_CURVE, show_display=False):
+    def get_public_node(self, n, ecdsa_curve_name=DEFAULT_CURVE, show_display=False, coin_name=None):
         n = self._convert_prime(n)
         if not ecdsa_curve_name:
-            ecdsa_curve_name=DEFAULT_CURVE
-        return self.call(proto.GetPublicKey(address_n=n, ecdsa_curve_name=ecdsa_curve_name, show_display=show_display))
+            ecdsa_curve_name = DEFAULT_CURVE
+        return self.call(proto.GetPublicKey(address_n=n, ecdsa_curve_name=ecdsa_curve_name, show_display=show_display, coin_name=coin_name))
 
     @field('address')
     @expect(proto.Address)
-    def get_address(self, coin_name, n, show_display=False, multisig=None):
+    def get_address(self, coin_name, n, show_display=False, multisig=None, script_type=types.SPENDADDRESS):
         n = self._convert_prime(n)
         if multisig:
-            return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display, multisig=multisig))
+            return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display, multisig=multisig, script_type=script_type))
         else:
-            return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display))
+            return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display, script_type=script_type))
 
     @field('address')
     @expect(proto.EthereumAddress)
@@ -423,33 +535,42 @@ class ProtocolMixin(object):
         n = self._convert_prime(n)
         return self.call(proto.EthereumGetAddress(address_n=n, show_display=show_display))
 
-    def ethereum_sign_tx(self, n, nonce, gas_price, gas_limit, to, value, data=''):
-        from rlp.utils import int_to_big_endian
+    @session
+    def ethereum_sign_tx(self, n, nonce, gas_price, gas_limit, to, value, data=None, chain_id=None):
+        def int_to_big_endian(value):
+            import rlp.utils
+            if value == 0:
+                return b''
+            return rlp.utils.int_to_big_endian(value)
 
         n = self._convert_prime(n)
 
-        try:
-            self.transport.session_begin()
+        msg = proto.EthereumSignTx(
+            address_n=n,
+            nonce=int_to_big_endian(nonce),
+            gas_price=int_to_big_endian(gas_price),
+            gas_limit=int_to_big_endian(gas_limit),
+            value=int_to_big_endian(value))
 
-            response = self.call(proto.EthereumSignTx(
-                address_n=n,
-                nonce=int_to_big_endian(nonce),
-                gas_price=int_to_big_endian(gas_price),
-                gas_limit=int_to_big_endian(gas_limit),
-                to=to,
-                value=int_to_big_endian(value)))
-            if data:
-                data, chunk = data[1024:], data[:1024]
-                response.data_initial_chunk = chunk
-                response.data_length = len(data)
+        if to:
+            msg.to = to
 
-            while response.HasField('data_length'):
-                data, chunk = data[1024:], data[:1024]
-                response = self.call(proto.EthereumTxAck(data_chunk=chunk))
+        if data:
+            msg.data_length = len(data)
+            data, chunk = data[1024:], data[:1024]
+            msg.data_initial_chunk = chunk
 
-            return response.signature_v, response.signature_r, response.signature_s
-        finally:
-            self.transport.session_end()
+        if chain_id:
+            msg.chain_id = chain_id
+
+        response = self.call(msg)
+
+        while response.HasField('data_length'):
+            data_length = response.data_length
+            data, chunk = data[data_length:], data[:data_length]
+            response = self.call(proto.EthereumTxAck(data_chunk=chunk))
+
+        return response.signature_v, response.signature_r, response.signature_s
 
     @field('entropy')
     @expect(proto.Entropy)
@@ -472,13 +593,13 @@ class ProtocolMixin(object):
     @expect(proto.Success)
     def apply_settings(self, label=None, language=None, use_passphrase=None, homescreen=None):
         settings = proto.ApplySettings()
-        if label != None:
+        if label is not None:
             settings.label = label
         if language:
             settings.language = language
-        if use_passphrase != None:
+        if use_passphrase is not None:
             settings.use_passphrase = use_passphrase
-        if homescreen != None:
+        if homescreen is not None:
             settings.homescreen = homescreen
 
         out = self.call(settings)
@@ -501,7 +622,7 @@ class ProtocolMixin(object):
     def sign_message(self, coin_name, n, message):
         n = self._convert_prime(n)
         # Convert message to UTF8 NFC (seems to be a bitcoin-qt standard)
-        message = normalize_nfc(message)
+        message = normalize_nfc(message).encode("utf-8")
         return self.call(proto.SignMessage(coin_name=coin_name, address_n=n, message=message))
 
     @expect(proto.SignedIdentity)
@@ -512,14 +633,17 @@ class ProtocolMixin(object):
     def get_ecdh_session_key(self, identity, peer_public_key, ecdsa_curve_name=DEFAULT_CURVE):
         return self.call(proto.GetECDHSessionKey(identity=identity, peer_public_key=peer_public_key, ecdsa_curve_name=ecdsa_curve_name))
 
-    def verify_message(self, address, signature, message):
+    @field('message')
+    @expect(proto.Success)
+    def set_u2f_counter(self, u2f_counter):
+        ret = self.call(proto.SetU2FCounter(u2f_counter=u2f_counter))
+        return ret
+
+    def verify_message(self, coin_name, address, signature, message):
         # Convert message to UTF8 NFC (seems to be a bitcoin-qt standard)
-        message = normalize_nfc(message)
+        message = normalize_nfc(message).encode("utf-8")
         try:
-            if address:
-                resp = self.call(proto.VerifyMessage(address=address, signature=signature, message=message))
-            else:
-                resp = self.call(proto.VerifyMessage(signature=signature, message=message))
+            resp = self.call(proto.VerifyMessage(address=address, signature=signature, message=message, coin_name=coin_name))
         except CallException as e:
             resp = e
         if isinstance(resp, proto.Success):
@@ -541,7 +665,7 @@ class ProtocolMixin(object):
 
     @field('value')
     @expect(proto.CipheredKeyValue)
-    def encrypt_keyvalue(self, n, key, value, ask_on_encrypt=True, ask_on_decrypt=True, iv=None):
+    def encrypt_keyvalue(self, n, key, value, ask_on_encrypt=True, ask_on_decrypt=True, iv=b''):
         n = self._convert_prime(n)
         return self.call(proto.CipherKeyValue(address_n=n,
                                               key=key,
@@ -549,11 +673,11 @@ class ProtocolMixin(object):
                                               encrypt=True,
                                               ask_on_encrypt=ask_on_encrypt,
                                               ask_on_decrypt=ask_on_decrypt,
-                                              iv=iv if iv is not None else ''))
+                                              iv=iv))
 
     @field('value')
     @expect(proto.CipheredKeyValue)
-    def decrypt_keyvalue(self, n, key, value, ask_on_encrypt=True, ask_on_decrypt=True, iv=None):
+    def decrypt_keyvalue(self, n, key, value, ask_on_encrypt=True, ask_on_decrypt=True, iv=b''):
         n = self._convert_prime(n)
         return self.call(proto.CipherKeyValue(address_n=n,
                                               key=key,
@@ -561,7 +685,7 @@ class ProtocolMixin(object):
                                               encrypt=False,
                                               ask_on_encrypt=ask_on_encrypt,
                                               ask_on_decrypt=ask_on_decrypt,
-                                              iv=iv if iv is not None else ''))
+                                              iv=iv))
 
     @field('tx_size')
     @expect(proto.TxSize)
@@ -585,7 +709,7 @@ class ProtocolMixin(object):
 
             tx = msg.transactions.add()
             if self.tx_api:
-                tx.CopyFrom(self.tx_api.get_tx(binascii.hexlify(inp.prev_hash)))
+                tx.CopyFrom(self.tx_api.get_tx(binascii.hexlify(inp.prev_hash).decode('utf-8')))
             else:
                 raise Exception('TX_API not defined')
             known_hashes.append(inp.prev_hash)
@@ -602,7 +726,7 @@ class ProtocolMixin(object):
         tx.outputs.extend(outputs)
 
         txes = {}
-        txes[''] = tx
+        txes[b''] = tx
 
         known_hashes = []
         for inp in inputs:
@@ -610,101 +734,108 @@ class ProtocolMixin(object):
                 continue
 
             if self.tx_api:
-                txes[inp.prev_hash] = self.tx_api.get_tx(binascii.hexlify(inp.prev_hash))
+                txes[inp.prev_hash] = self.tx_api.get_tx(binascii.hexlify(inp.prev_hash).decode('utf-8'))
             else:
                 raise Exception('TX_API not defined')
             known_hashes.append(inp.prev_hash)
 
         return txes
 
-    def sign_tx(self, coin_name, inputs, outputs, debug_processor=None):
+    @session
+    def sign_tx(self, coin_name, inputs, outputs, version=None, lock_time=None, debug_processor=None):
 
         start = time.time()
         txes = self._prepare_sign_tx(coin_name, inputs, outputs)
 
-        try:
-            self.transport.session_begin()
+        # Prepare and send initial message
+        tx = proto.SignTx()
+        tx.inputs_count = len(inputs)
+        tx.outputs_count = len(outputs)
+        tx.coin_name = coin_name
+        if version is not None:
+            tx.version = version
+        if lock_time is not None:
+            tx.lock_time = lock_time
+        res = self.call(tx)
 
-            # Prepare and send initial message
-            tx = proto.SignTx()
-            tx.inputs_count = len(inputs)
-            tx.outputs_count = len(outputs)
-            tx.coin_name = coin_name
-            res = self.call(tx)
+        # Prepare structure for signatures
+        signatures = [None] * len(inputs)
+        serialized_tx = b''
 
-            # Prepare structure for signatures
-            signatures = [None] * len(inputs)
-            serialized_tx = ''
+        counter = 0
+        while True:
+            counter += 1
 
-            counter = 0
-            while True:
-                counter += 1
+            if isinstance(res, proto.Failure):
+                raise CallException("Signing failed")
 
-                if isinstance(res, proto.Failure):
-                    raise CallException("Signing failed")
+            if not isinstance(res, proto.TxRequest):
+                raise CallException("Unexpected message")
 
-                if not isinstance(res, proto.TxRequest):
-                    raise CallException("Unexpected message")
+            # If there's some part of signed transaction, let's add it
+            if res.HasField('serialized') and res.serialized.HasField('serialized_tx'):
+                log("RECEIVED PART OF SERIALIZED TX (%d BYTES)" % len(res.serialized.serialized_tx))
+                serialized_tx += res.serialized.serialized_tx
 
-                # If there's some part of signed transaction, let's add it
-                if res.HasField('serialized') and res.serialized.HasField('serialized_tx'):
-                    log("RECEIVED PART OF SERIALIZED TX (%d BYTES)" % len(res.serialized.serialized_tx))
-                    serialized_tx += res.serialized.serialized_tx
+            if res.HasField('serialized') and res.serialized.HasField('signature_index'):
+                if signatures[res.serialized.signature_index] is not None:
+                    raise Exception("Signature for index %d already filled" % res.serialized.signature_index)
+                signatures[res.serialized.signature_index] = res.serialized.signature
 
-                if res.HasField('serialized') and res.serialized.HasField('signature_index'):
-                    if signatures[res.serialized.signature_index] != None:
-                        raise Exception("Signature for index %d already filled" % res.serialized.signature_index)
-                    signatures[res.serialized.signature_index] = res.serialized.signature
+            if res.request_type == types.TXFINISHED:
+                # Device didn't ask for more information, finish workflow
+                break
 
-                if res.request_type == types.TXFINISHED:
-                    # Device didn't ask for more information, finish workflow
-                    break
+            # Device asked for one more information, let's process it.
+            current_tx = txes[res.details.tx_hash]
 
-                # Device asked for one more information, let's process it.
-                current_tx = txes[res.details.tx_hash]
+            if res.request_type == types.TXMETA:
+                msg = types.TransactionType()
+                msg.version = current_tx.version
+                msg.lock_time = current_tx.lock_time
+                msg.inputs_cnt = len(current_tx.inputs)
+                if res.details.tx_hash:
+                    msg.outputs_cnt = len(current_tx.bin_outputs)
+                else:
+                    msg.outputs_cnt = len(current_tx.outputs)
+                msg.extra_data_len = len(current_tx.extra_data)
+                res = self.call(proto.TxAck(tx=msg))
+                continue
 
-                if res.request_type == types.TXMETA:
-                    msg = types.TransactionType()
-                    msg.version = current_tx.version
-                    msg.lock_time = current_tx.lock_time
-                    msg.inputs_cnt = len(current_tx.inputs)
-                    if res.details.tx_hash:
-                        msg.outputs_cnt = len(current_tx.bin_outputs)
-                    else:
-                        msg.outputs_cnt = len(current_tx.outputs)
-                    res = self.call(proto.TxAck(tx=msg))
-                    continue
+            elif res.request_type == types.TXINPUT:
+                msg = types.TransactionType()
+                msg.inputs.extend([current_tx.inputs[res.details.request_index], ])
+                res = self.call(proto.TxAck(tx=msg))
+                continue
 
-                elif res.request_type == types.TXINPUT:
-                    msg = types.TransactionType()
-                    msg.inputs.extend([current_tx.inputs[res.details.request_index], ])
-                    res = self.call(proto.TxAck(tx=msg))
-                    continue
+            elif res.request_type == types.TXOUTPUT:
+                msg = types.TransactionType()
+                if res.details.tx_hash:
+                    msg.bin_outputs.extend([current_tx.bin_outputs[res.details.request_index], ])
+                else:
+                    msg.outputs.extend([current_tx.outputs[res.details.request_index], ])
 
-                elif res.request_type == types.TXOUTPUT:
-                    msg = types.TransactionType()
-                    if res.details.tx_hash:
-                        msg.bin_outputs.extend([current_tx.bin_outputs[res.details.request_index], ])
-                    else:
-                        msg.outputs.extend([current_tx.outputs[res.details.request_index], ])
+                if debug_processor is not None:
+                    # If debug_processor function is provided,
+                    # pass thru it the request and prepared response.
+                    # This is useful for unit tests, see test_msg_signtx
+                    msg = debug_processor(res, msg)
 
-                    if debug_processor != None:
-                        # If debug_processor function is provided,
-                        # pass thru it the request and prepared response.
-                        # This is useful for unit tests, see test_msg_signtx
-                        msg = debug_processor(res, msg)
+                res = self.call(proto.TxAck(tx=msg))
+                continue
 
-                    res = self.call(proto.TxAck(tx=msg))
-                    continue
-
-        finally:
-            self.transport.session_end()
+            elif res.request_type == types.TXEXTRADATA:
+                o, l = res.details.extra_data_offset, res.details.extra_data_len
+                msg = types.TransactionType()
+                msg.extra_data = current_tx.extra_data[o:o + l]
+                res = self.call(proto.TxAck(tx=msg))
+                continue
 
         if None in signatures:
             raise Exception("Some signatures are missing!")
 
-        log("SIGNED IN %.03f SECONDS, CALLED %d MESSAGES, %d BYTES" % \
-                (time.time() - start, counter, len(serialized_tx)))
+        log("SIGNED IN %.03f SECONDS, CALLED %d MESSAGES, %d BYTES" %
+            (time.time() - start, counter, len(serialized_tx)))
 
         return (signatures, serialized_tx)
 
@@ -717,60 +848,83 @@ class ProtocolMixin(object):
 
     @field('message')
     @expect(proto.Success)
-    def recovery_device(self, word_count, passphrase_protection, pin_protection, label, language):
-        if self.features.initialized:
+    def recovery_device(self, word_count, passphrase_protection, pin_protection, label, language, type=types.RecoveryDeviceType_ScrambledWords, expand=False, dry_run=False):
+        if self.features.initialized and not dry_run:
             raise Exception("Device is initialized already. Call wipe_device() and try again.")
 
         if word_count not in (12, 18, 24):
             raise Exception("Invalid word count. Use 12/18/24")
 
-        res = self.call(proto.RecoveryDevice(word_count=int(word_count),
-                                   passphrase_protection=bool(passphrase_protection),
-                                   pin_protection=bool(pin_protection),
-                                   label=label,
-                                   language=language,
-                                   enforce_wordlist=True))
+        self.recovery_matrix_first_pass = True
+
+        self.expand = expand
+        if self.expand:
+            # optimization to load the wordlist once, instead of for each recovery word
+            self.mnemonic_wordlist = Mnemonic('english')
+
+        res = self.call(proto.RecoveryDevice(
+            word_count=int(word_count),
+            passphrase_protection=bool(passphrase_protection),
+            pin_protection=bool(pin_protection),
+            label=label,
+            language=language,
+            enforce_wordlist=True,
+            type=type,
+            dry_run=dry_run))
 
         self.init_device()
         return res
 
     @field('message')
     @expect(proto.Success)
-    def reset_device(self, display_random, strength, passphrase_protection, pin_protection, label, language):
+    @session
+    def reset_device(self, display_random, strength, passphrase_protection, pin_protection, label, language, u2f_counter, skip_backup):
         if self.features.initialized:
             raise Exception("Device is initialized already. Call wipe_device() and try again.")
 
         # Begin with device reset workflow
         msg = proto.ResetDevice(display_random=display_random,
                                 strength=strength,
-                                language=language,
                                 passphrase_protection=bool(passphrase_protection),
                                 pin_protection=bool(pin_protection),
-                                label=label)
+                                language=language,
+                                label=label,
+                                u2f_counter=u2f_counter,
+                                skip_backup=bool(skip_backup))
 
         resp = self.call(msg)
         if not isinstance(resp, proto.EntropyRequest):
             raise Exception("Invalid response, expected EntropyRequest")
 
         external_entropy = self._get_local_entropy()
-        log("Computer generated entropy: " + binascii.hexlify(external_entropy))
+        log("Computer generated entropy: " + binascii.hexlify(external_entropy).decode('ascii'))
         ret = self.call(proto.EntropyAck(entropy=external_entropy))
         self.init_device()
         return ret
 
     @field('message')
     @expect(proto.Success)
-    def load_device_by_mnemonic(self, mnemonic, pin, passphrase_protection, label, language, skip_checksum=False):
-        m = Mnemonic('english')
-        if not skip_checksum and not m.check(mnemonic):
-            raise Exception("Invalid mnemonic checksum")
+    def backup_device(self):
+        ret = self.call(proto.BackupDevice())
+        return ret
 
+    @field('message')
+    @expect(proto.Success)
+    def load_device_by_mnemonic(self, mnemonic, pin, passphrase_protection, label, language, skip_checksum=False, expand=False):
         # Convert mnemonic to UTF8 NKFD
         mnemonic = Mnemonic.normalize_string(mnemonic)
 
         # Convert mnemonic to ASCII stream
         mnemonic = normalize_nfc(mnemonic)
 
+        m = Mnemonic('english')
+
+        if expand:
+            mnemonic = m.expand(mnemonic)
+
+        if not skip_checksum and not m.check(mnemonic):
+            raise Exception("Invalid mnemonic checksum")
+
         if self.features.initialized:
             raise Exception("Device is initialized already. Call wipe_device() and try again.")
 
@@ -797,10 +951,10 @@ class ProtocolMixin(object):
         node = types.HDNodeType()
         data = binascii.hexlify(tools.b58decode(xprv, None))
 
-        if data[90:92] != '00':
+        if data[90:92] != b'00':
             raise Exception("Contain invalid private key")
 
-        checksum = hashlib.sha256(hashlib.sha256(binascii.unhexlify(data[:156])).digest()).hexdigest()[:8]
+        checksum = binascii.hexlify(hashlib.sha256(hashlib.sha256(binascii.unhexlify(data[:156])).digest()).digest()[:4])
         if checksum != data[156:]:
             raise Exception("Checksum doesn't match")
 
@@ -815,8 +969,8 @@ class ProtocolMixin(object):
         node.depth = int(data[8:10], 16)
         node.fingerprint = int(data[10:18], 16)
         node.child_num = int(data[18:26], 16)
-        node.chain_code = data[26:90].decode('hex')
-        node.private_key = data[92:156].decode('hex')  # skip 0x00 indicating privkey
+        node.chain_code = binascii.unhexlify(data[26:90])
+        node.private_key = binascii.unhexlify(data[92:156])  # skip 0x00 indicating privkey
 
         resp = self.call(proto.LoadDevice(node=node,
                                           pin=pin,
@@ -826,32 +980,61 @@ class ProtocolMixin(object):
         self.init_device()
         return resp
 
+    @session
     def firmware_update(self, fp):
-        if self.features.bootloader_mode == False:
+        if self.features.bootloader_mode is False:
             raise Exception("Device must be in bootloader mode")
 
-        resp = self.call(proto.FirmwareErase())
+        data = fp.read()
+
+        resp = self.call(proto.FirmwareErase(length=len(data)))
         if isinstance(resp, proto.Failure) and resp.code == types.Failure_FirmwareError:
             return False
 
-        data = fp.read()
-        fingerprint = hashlib.sha256(data[256:]).hexdigest()
-        log("Firmware fingerprint: " + fingerprint)
-        resp = self.call(proto.FirmwareUpload(payload=data))
-
+        # TREZORv1 method
         if isinstance(resp, proto.Success):
-            return True
+            fingerprint = hashlib.sha256(data[256:]).hexdigest()
+            log("Firmware fingerprint: " + fingerprint)
+            resp = self.call(proto.FirmwareUpload(payload=data))
+            if isinstance(resp, proto.Success):
+                return True
+            elif isinstance(resp, proto.Failure) and resp.code == types.Failure_FirmwareError:
+                return False
+            raise Exception("Unexpected result %s" % resp)
 
-        elif isinstance(resp, proto.Failure) and resp.code == types.Failure_FirmwareError:
-            return False
+        # TREZORv2 method
+        if isinstance(resp, proto.FirmwareRequest):
+            import pyblake2
+            while True:
+                payload = data[resp.offset:resp.offset + resp.length]
+                digest = pyblake2.blake2s(payload).digest()
+                resp = self.call(proto.FirmwareUpload(payload=payload, hash=digest))
+                if isinstance(resp, proto.FirmwareRequest):
+                    continue
+                elif isinstance(resp, proto.Success):
+                    return True
+                elif isinstance(resp, proto.Failure) and resp.code == types.Failure_FirmwareError:
+                    return False
+                raise Exception("Unexpected result %s" % resp)
+
+        raise Exception("Unexpected message %s" % resp)
+
+    @field('message')
+    @expect(proto.Success)
+    def self_test(self):
+        if self.features.bootloader_mode is False:
+            raise Exception("Device must be in bootloader mode")
+
+        return self.call(proto.SelfTest())
 
-        raise Exception("Unexpected result %s" % resp)
 
 class TrezorClient(ProtocolMixin, TextUIMixin, BaseClient):
     pass
 
+
 class TrezorClientDebug(ProtocolMixin, TextUIMixin, DebugWireMixin, BaseClient):
     pass
 
+
 class TrezorDebugClient(ProtocolMixin, DebugLinkMixin, DebugWireMixin, BaseClient):
     pass
