@@ -1,18 +1,41 @@
+# This file is part of the TREZOR project.
+#
+# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
+# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
+# Copyright (C) 2016      Jochen Hoenicke <hoenicke@gmail.com>
+#
+# This library is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Lesser General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This library is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Lesser General Public License for more details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with this library.  If not, see <http://www.gnu.org/licenses/>.
+
 '''BridgeTransport implements transport TREZOR Bridge (aka trezord).'''
 
 import json
 import requests
 from . import protobuf_json
 from . import messages_pb2 as proto
-from .transport import Transport
+from .transport import TransportV1
 
 TREZORD_HOST = 'https://localback.net:21324'
-CONFIG_URL = 'https://mytrezor.com/data/plugin/config_signed.bin'
+CONFIG_URL = 'https://wallet.trezor.io/data/config_signed.bin'
+
 
 def get_error(resp):
     return ' (error=%d str=%s)' % (resp.status_code, resp.json()['error'])
 
-class BridgeTransport(Transport):
+
+class BridgeTransport(TransportV1):
+    CONFIGURED = False
+
     def __init__(self, device, *args, **kwargs):
         self.configure()
 
@@ -26,7 +49,9 @@ class BridgeTransport(Transport):
 
     @staticmethod
     def configure():
-        r = requests.get(CONFIG_URL)
+        if BridgeTransport.CONFIGURED:
+            return
+        r = requests.get(CONFIG_URL, verify=False)
         if r.status_code != 200:
             raise Exception('Could not fetch config from %s' % CONFIG_URL)
 
@@ -35,13 +60,13 @@ class BridgeTransport(Transport):
         r = requests.post(TREZORD_HOST + '/configure', data=config)
         if r.status_code != 200:
             raise Exception('trezord: Could not configure' + get_error(r))
+        BridgeTransport.CONFIGURED = True
 
     @classmethod
     def enumerate(cls):
         """
         Return a list of available TREZOR devices.
         """
-        devices = {}
         cls.configure()
         r = requests.get(TREZORD_HOST + '/enumerate')
         if r.status_code != 200:
@@ -51,7 +76,6 @@ class BridgeTransport(Transport):
 
         return enum
 
-
     def _open(self):
         r = self.conn.post(TREZORD_HOST + '/acquire/%s' % self.path)
         if r.status_code != 200:
@@ -66,10 +90,12 @@ class BridgeTransport(Transport):
         else:
             self.session = None
 
-    def ready_to_read(self):
-        return self.response != None
+    def _ready_to_read(self):
+        return self.response is not None
 
-    def _write(self, msg, protobuf_msg):
+    def write(self, protobuf_msg):
+        # Override main 'write' method, HTTP transport cannot be
+        # splitted to chunks
         cls = protobuf_msg.__class__.__name__
         msg = protobuf_json.pb2json(protobuf_msg)
         payload = '{"type": "%s", "message": %s}' % (cls, json.dumps(msg))
@@ -85,4 +111,4 @@ class BridgeTransport(Transport):
         cls = getattr(proto, self.response['type'])
         inst = cls()
         pb = protobuf_json.json2pb(inst, self.response['message'])
-        return ('protobuf', pb)
+        return (0, 'protobuf', pb)
